"""Static settings"""

DEVICE_TYPES = [
    "cisco_asr",
    "cisco_ios",
    "cisco_nxos",
    "hp_comware",
    "hp_procurve",
]
RANGE_SPLITTER = "-"
RANGE_SPLITTER_HPE = " to "
RANGE_SPLITTERS = [RANGE_SPLITTER, RANGE_SPLITTER_HPE]
SPLITTER = ","
SPLITTER_HPE = " "

BRIEF_ALL_I = -1
BRIEF_ALL_S = "-1"
